import math
import matplotlib.pyplot as plt
import numpy as np
from sklearn.neighbors import KDTree
import glob
from tqdm import tqdm
from os.path import exists, join


class urban_object:
    """
    Define an urban object
    """
    def __init__(self, filenm):
        """
        Initialize the object
        """
        # obtain the cloud name
        self.cloud_name = filenm.split('/\\')[-1][-7:-4]

        # obtain the cloud ID
        self.cloud_ID = int(self.cloud_name)

        # obtain the label
        self.label = math.floor(1.0*self.cloud_ID/100)

        # obtain the points
        self.points = read_xyz(filenm)

        # initialize the feature vector
        self.feature = []

    def compute_features(self):
        """
        Compute the features, here we provide two example features. You're encouraged to design your own features
        """
        # calculate the height
        height = np.amax(self.points[:, 2])
        self.feature.append(height)

        # get the root point
        root = self.points[[np.argmin(self.points[:, 2])]]

        # construct the 2D kd tree
        kd_tree_2d = KDTree(self.points[:, :2], leaf_size=5)

        # compute the root point planar density
        radius_root = 0.2
        count = kd_tree_2d.query_radius(root[:, :2], r=radius_root, count_only=True)
        root_density = 1.0*count[0] / len(self.points)
        self.feature.append(root_density)


def read_xyz(filenm):
    """
    Reading points
        filenm: the file name
    """
    points = []
    with open(filenm, 'r') as f_input:
        for line in f_input:
            p = line.split()
            p = [float(i) for i in p]
            points.append(p)
    points = np.array(points).astype(np.float32)
    return points


def feature_preparation(data_path):
    """
    Prepare features of the input point cloud objects
        data_path: the path to read data
    """
    # check if the current data file exist
    data_file = 'data.txt'
    if exists(data_file):
        return

    # obtain the files in the folder
    files = np.sort(glob.glob(join(data_path, '*.xyz')))

    # initialize the data
    input_data = []

    # get the overall point cloud file number
    N = len(files)

    # initialize the progress bar
    progress = tqdm(range(N))

    # retrieve each data object and obtain the feature vector
    for i in progress:
        # get the file
        file_name = files[i]

        # read data
        i_object = urban_object(filenm=file_name)

        # calculate features
        i_object.compute_features()

        # add the data to the list
        i_data = [i_object.cloud_ID, i_object.label] + i_object.feature
        input_data += [i_data]

        # set progress description
        progress.set_description("Processed files: ")

    # transform the output data
    outputs = np.array(input_data).astype(np.float32)

    # write the output to a local file
    data_header = 'ID,label,height,root_density'
    np.savetxt(data_file, outputs, fmt='%10.5f', delimiter=',', newline='\n', header=data_header)


def data_loading(data_file='data.txt'):
    """
    Read the data with features from the data file
        data_file: the local file to read data with features and labels
    """
    # load data
    data = np.loadtxt(data_file, dtype=np.float32, delimiter=',', comments='#')

    # extract object ID, feature X and label Y
    ID = data[:, 0].astype(np.int32)
    Y = data[:, 1].astype(np.int32)
    X = data[:, 2:].astype(np.float32)

    return ID, X, Y


def feature_visualization(X):
    """
    Visualize the features
        X: input features. This assumes classes are stored in a sequential manner
    """
    # initialize a plot
    fig, ax = plt.subplots()
    plt.title("feature subset visualization of 5 classes", fontsize="small")

    # define the labels and corresponding colors
    colors = ['firebrick', 'grey', 'darkorange', 'dodgerblue', 'olivedrab']
    labels = ['building', 'car', 'fence', 'pole', 'tree']

    # plot the data with first two features
    for i in range(5):
        ax.scatter(X[100*i:100*(i+1), 0], X[100*i:100*(i+1), 1], marker="o", c=colors[i], edgecolor="k", label=labels[i])

    # show the figure with labels
    """
    Replace the axis labels with your own feature names
    """
    plt.xlabel('x1: height')
    plt.ylabel('x2: root density')
    ax.legend()
    plt.show()


def SVM_classification(X, Y):
    """
    Conduct SVM classification
        X: features
        Y: labels
    """
    print("TODO: implement SVM")
    pass


def RF_classification(X, Y):
    """
    Conduct RF classification
        X: features
        Y: labels
    """
    print("TODO: implement RF")
    pass


def Evaluation(Y_preds=None, Y_true=None):
    """
    Evaluate the performance
        Y_preds: predicted labels
        Y_true: true labels
    """
    print("TODO: implement evaluation")
    pass


if __name__=='__main__':
    # specify the data folder
    """"Here you need to specify your own path"""
    path = '/Users/lnan/Downloads/pointclouds-500'

    # conduct feature preparation
    print('Start preparing features')
    feature_preparation(data_path=path)

    # load the data
    print('Start loading data from the local file')
    ID, X, Y = data_loading()

    # visualize features
    print('Visualize the features')
    feature_visualization(X=X)

    # SVM classification
    print('Start SVM classification')
    SVM_classification(X, Y)

    # RF classification
    print('Start RF classification')
    RF_classification(X, Y)

    # Evaluate results
    print('Start evaluating the result')
    Evaluation()

